/***************************************************************************

	akc6951.c

	control for AKC6951 dsp radio

	2013/11/19 for SBDBT32 (PIC32MX695F512H)
***************************************************************************/
#include <stdio.h>
#include "HardwareProfile.h"
#include "akc6951.h"
#include "pic_delay.h"

/*--------------------------------------------------------------------------
	AKC6951 I2C Register definition
--------------------------------------------------------------------------*/
#if 0

// AKC6951 setting value
//  get from read/write register(address from 0 to 13)
#define power_on    ((RegVal[0]&0x8)>>7)    //v4
#define fm_en       ((RegVal[0]&0x40)>>6)   //v4
#define tune        ((RegVal[0]&0x20)>>5)   //v4
#define seek        ((RegVal[0]&0x10)>>4)   //v4
#define seekup      ((RegVal[0]&0x08)>>3)   //v4
#define mute        ((RegVal[0]&0x04)>>2)
#define amband      ((RegVal[1]&0xf8)>>3)
#define fmband       (RegVal[1]&0x07) 
#define ref_32kmode ((RegVal[2]&0x40)>>6)
#define mode3k      ((RegVal[2]&0x20)>>5)
#define chan        ((RegVal[2]&0x001f)<<8|RegVal[3])
#define usr_chan_start (RegVal[4])
#define usr_chan_stop  (RegVal[5])
#define volume      ((RegVal[6]&0xFC)>>2)
#define line        ((RegVal[6]&0x02)>>1)
#define phase_inv    (RegVal[6]&0x01)
#define De          ((RegVal[7]&0x20)>>5)
#define BBEN        ((RegVal[7]&0x10)>>4)
#define Stereo      ((RegVal[7]&0x08)>>3)
#define Mono        ((RegVal[7]&0x04)>>2)
#define bw           (RegVal[7]&0x03)
#define fm_cnr_th   ((RegVal[8]&0xc0)>>6)
#define am_cnr_th   ((RegVal[8]&0x30)>>4)
#define fd_th       ((RegVal[8]&0x0c)>>2)
#define streo_th     (RegVal[8]&0x03)
#define pd_adc_vol  ((RegVal[9]&0x08)>>3)
#define osc_en      ((RegVal[9]&0x04)>>2)
#define lv_en        (RegVal[9]&0x01)
#define space       ((RegVal[11]&0x30)>>4)
#define pd_adc      ((RegVal[12]&0x80)>>7)
#define pd_rx       ((RegVal[12]&0x20)>>5)
#define st_led      ((RegVal[13]&0x40)>>6)
#define vol_pre     ((RegVal[13]&0x0c)>>2)

#endif // if 0, comment out

// AKC6951 status value 
//   get from read only register(address from 20 to 27)
#define ST          ((RegVal[20]&0x80)>>7)
#define STC         ((RegVal[20]&0x40)>>6)
#define TUNED       ((RegVal[20]&0x20)>>5)
#define READ_CHAN   ((RegVal[20]&0x1f)<<8)+(RegVal[21]&0xff)
#define MODE3K_F    ((RegVal[22]&0x80)>>7)
#define CNR_AM       (RegVal[22]&0x7f)
#define ST_DEM      ((RegVal[23]&0x80)>>7)
#define CNR_FM       (RegVal[23]&0x7f)
#define PGALEVEL_RF ((RegVal[24]&0xe0)>>5)
#define PGALEVEL_IF ((RegVal[24]&0x1c)>>2)
#define FM_MODE     ((RegVal[24]&0x02)>>1)  //v4
#define LVMODE       (RegVal[24]&0x01)
#define VBAT         (RegVal[25]&0x3f)
#define FD_NUM        RegVal[26]
#define RSSI         (RegVal[27]&0x7f)

// AM Band (amband)
#define LW  0b00000	// LW:150-285KHz, 3KHz Step
#define MW1 0b00001 // MW1:520-1710KHz, 5KHz Step
#define MW2 0b00010 // MW2:522-1620KHz, 9KHz Step (Japan)
#define MW3 0b00011 // MW3:520-1710KHz, 10KHz Step
#define MW4 0b10010 // MW4:520-1730KHz, 5KHz Step

// FM Band (fmband)
#define FM1 0b000  // FM1:87-108MHz
#define FM2 0b001  // FM2:76-108MHz
#define FM3 0b010  // FM3:70-93MHz
#define FM4 0b011  // FM4:76-90MHz (Japan)
#define FM5 0b100  // FM5:64-88MHz
#define TV1 0b101  // TV1:56.25-91.75MHz   (Low Channel)
#define TV2 0b110  // TV2:174.75-222.25MHz (High Channel)

// FM Band Width (bw) FM channel spacing
#define BW150K    0
#define BW200K    1
#define BW50K     2
#define BW100K    3

// Setting of FM seek step
#define STEP25K   0
#define STEP50K   1
#define STEP100K  2
#define STEP200K  3

// Setting volume of output
#define VL0DB     0
#define VL3D5DB   1
#define VL7DB     2
#define VL10D5DB  3

// Register 0
#define	PowerOn()      {RegVal[0]|=0x80;}
#define PowerOff()     {RegVal[0]&=0x7f;}
#define FMmode()       {RegVal[0]|=0x40;}
#define AMmode()       {RegVal[0]&=0xbf;}
#define TrigTune()     {RegVal[0]|=0x20;}
#define ResetTune()    {RegVal[0]&=0xdf;}
#define TrigSeek()     {RegVal[0]|=0x10;}
#define ResetSeek()    {RegVal[0]&=0xef;}
#define SeekUp()       {RegVal[0]|=0x08;}
#define SeekDown()     {RegVal[0]&=0xf7;}
#define MuteOn()       {RegVal[0]|=0x04;}
#define MuteOff()      {RegVal[0]&=0xfb;}

// Register 1
#define AMBand(x)      {RegVal[1]=((RegVal[1]&0x07)|((x) << 3));}
#define FMBand(x)      {RegVal[1]=((RegVal[1]&0xf8)|(x));}

// Register 2,3
#define OSC32K()       {RegVal[2]|=0x40;}
#define OSC12M()       {RegVal[2]&=0xbf;}
#define AM3KStep()     {RegVal[2]|=0x20;}
#define AM5KStep()     {RegVal[2]&=0xdf;}
#define SetChannel(x)  {RegVal[2]=((RegVal[2]&0xe0)|(((x)>>8)&0x01f));RegVal[3]=((x)&0xff);}

// Register 4
#define UsrChanStart(x) {RegVal[4]=(x);}

// Register 5
#define UsrChanStop(x)  {RegVal[5]=(x);}

// Register 6
#define Volume(x)       {RegVal[6]=((RegVal[6]&0x3)|((x)<<2));}
#define LineInMode()    {RegVal[6]|=0x02;}
#define RadioMode()     {RegVal[6]&=0xfd;}
#define SingleSPK()     {RegVal[6]|=1;}
#define DualSPK()       {RegVal[6]&=0xfe;}

// Register 7
#define NormalOPR()     {RegVal[7]&=0x3f;}
#define De50()          {RegVal[7]|=0x20;} // Japan
#define De75()          {RegVal[7]&=0xdf;}
#define BaseBoostOn()   {RegVal[7]|=0x10;}
#define BaseBoostOff()  {RegVal[7]&=0xef;}
#define AutoMode()      {RegVal[7]&=0xf3;}
#define StereoMode()    {RegVal[7]&=0xf3;RegVal[7]|=0x08;}
#define MonoMode()      {RegVal[7]&=0xf3;RegVal[7]|=0x04;}
#define FMBandWidth(x)  {RegVal[7]&=0xfc;RegVal[7]|=(x);}

// Register 8  don't set, use default
#define FMCNR(x)        {RegVal[8]&=0x3f;RegVal[8]|=((x)<<6);}
#define AMCNR(x)        {RegVal[8]&=0xcf;RegVal[8]|=((x)<<4);}
#define FDTH(x)         {RegVal[8]&=0xf3;RegVal[8]|=((x)<<2);}
#define StereoTH(x)     {RegVal[8]&=0xfc;RegVal[8]|=(x);}

// Register 9 
#define VolCtrlByReg()  {RegVal[9]&=0xf7;}
#define VolCtrlByI2C()  {RegVal[9]|=0x08;}
#define SelEXO()        {RegVal[9]&=0xfb;}
#define SelXTL()        {RegVal[9]|=0x04;}
#define LowPowerOn()    {RegVal[9]|=1;}
#define LowPowerOff()   {RegVal[9]&=0xfe;}

// Register 10 reserve

// Register 11
#define FMSeekStep(x)   {RegVal[11]&=0xcf;RegVal[11]|=((x)<<4);}

// Register 12
#define ADCEnable()     {RegVal[12]&=0x7f;}
#define ADCDisable()    {RegVal[12]|=0x80;}
#define RFEnable()      {RegVal[12]&=0xdf;}
#define RFDisable()     {RegVal[12]|=0x20;}

// Register 13
#define TuneLED()       {RegVal[13]&=0xbf;}
#define FMStereoLED()   {RegVal[13]|=0x40;}
#define VolOut(x)       {RegVal[13]&=0xf3;RegVal[13]|=((x)<<2);}


/*--------------------------------------------------------------------------
	AKC6951 I2C device definition
--------------------------------------------------------------------------*/

#define	DEVADDR  0x20  // AKC6951 device address
#define	MW       0     // I2C master write
#define MR       1     // I2C master read


/*--------------------------------------------------------------------------
	function prototype
--------------------------------------------------------------------------*/

// AKC6951 register access
static char  akc6951_read_byte(char regadr);
static void  akc6951_read_byte_block(char regadr, char *data, int nbytes);
static void  akc6951_write_byte(char regadr, char data);
static void  akc6951_write_byte_block(char regadr, char *data, int nbytes);

// I2C handling
static void  i2cOpenStart(void);
// static void  i2cReStart(void);
static void  i2cStopClose(void);
static int   i2cSend(char data);
static char  i2cRec(void);
static void  i2cAckSend(void);
static void  i2cNAckSend(void);


//**************************************************************************
//**************************************************************************
//     Tuning
//**************************************************************************
//**************************************************************************
//
int Tuned(unsigned long freq)
{
	if(freq >= 76000 && freq <= 91000){  // Japan FM BAND
		return akc6951_FM_tunning(freq);
	}else if(freq >= 522 && freq <= 1620){ // Japan MW BAND
		return akc6951_AM_tunning(freq);
	}
	return TUNE_FAIL;
}


//**************************************************************************
//**************************************************************************
//     gloval variable
//**************************************************************************
//**************************************************************************

// AKC6951 REGISTER Value
char RegVal[28];


//**************************************************************************
//**************************************************************************
//     AKC6951 Handling
//**************************************************************************
//**************************************************************************
//
/*--------------------------------------------------------------------------
	akc6951_initialize
--------------------------------------------------------------------------*/

void akc6951_initialize(void)
{
	// get default value
	akc6951_read_byte_block(0, RegVal, 14);
	akc6951_read_byte_block(20, RegVal+20, 8);

	// set initial parameter
	PowerOn();
	FMmode();  // select FM
	SeekUp();
	MuteOff();

	AMBand(MW2); // 522-1620KHz
	FMBand(FM4); // 76-90MHz

	OSC32K();
	AM3KStep();
	SetChannel((77800-30000)/25); // 77.8MHz    77800KHz=25KHz*x+30000KHz

	UsrChanStart((76000-30000)/25/32); // 76MHz
	UsrChanStop((90000-30000)/25/32);  // 90MHz

	Volume(63);	
	RadioMode();
	DualSPK();
	//SingleSPK();

	NormalOPR();
	De50();
	BaseBoostOff();
	StereoMode();
	FMBandWidth(BW50K);

	VolCtrlByI2C();
	SelXTL();
	LowPowerOn();

	FMSeekStep(STEP50K);

	ADCEnable();
	RFEnable();

	TuneLED();
	VolOut(VL0DB);


	// write parameter
	akc6951_write_byte_block(1, RegVal+1, 13);

	// write PowerON, FMmode 
	akc6951_write_byte(0, RegVal[0]);

}

/*--------------------------------------------------------------------------
	akc6951_register_dump
--------------------------------------------------------------------------*/

void akc6951_register_dump(void)
{
	int n;

	// get value
	akc6951_read_byte_block(0, RegVal, 14);
	akc6951_read_byte_block(20, RegVal+20, 8);

	for(n = 0; n < 14; n++){
		printf("Register[%2d]=0x%02x\n", n, RegVal[n] & 0xff);
	}

	for(n = 20; n < 28; n++){
		printf("Register[%2d]=0x%02x\n", n, RegVal[n] & 0xff);
	}
}


/*--------------------------------------------------------------------------
	akc6951_VOL_setting
--------------------------------------------------------------------------*/

int akc6951_VOL_setting(int vol)
{
	printf("akc6951: Volume setting, val=%d\n", vol);

	// set volume 
	Volume(vol & 0xff);	
	akc6951_write_byte(6, RegVal[6]);	// Register6

	return 0;
}

/*--------------------------------------------------------------------------
	akc6951_BaseBoost(int sw)
--------------------------------------------------------------------------*/

int akc6951_BaseBoost(int sw)
{
	printf("akc6951: Base Boost %s.\n", sw ? "ON": "OFF");

	// set Base Boost
	if(sw){
		BaseBoostOn();
	}else{
		BaseBoostOff();
	}
	akc6951_write_byte(7, RegVal[7]);	// Register7
	return 0;
}

/*--------------------------------------------------------------------------
	akc6951_SelectSpeaker(int sw)
--------------------------------------------------------------------------*/

int akc6951_SelectSpeaker(int sw)
{
	printf("akc6951: Select Speaker %s.\n", sw ? "Dual": "Single");

	// select speaker
	if(sw){
		DualSPK();
	}else{
		SingleSPK();
	}
	akc6951_write_byte(6, RegVal[6]);	// Register6
	return 0;
}

/*--------------------------------------------------------------------------
	akc6951_FM_tunning
	freq KHZ 76000-90000
--------------------------------------------------------------------------*/

int akc6951_FM_tunning(long freq)
{
	short   channel;
	unsigned short  to;

	printf("akc6951: FM tuning, freq=%ld\n", freq);

	// set request channel
	channel = (short)((freq - 30000L)/25L);	
	SetChannel(channel);
	akc6951_write_byte_block(2, RegVal+2, 2);	// Register2,Register3

	// trigger tunning
	FMmode();     // select FM
	ResetTune();
	akc6951_write_byte(0, RegVal[0]);	// Register0
	TrigTune();
	akc6951_write_byte(0, RegVal[0]);	// Register0

	// wait for tunning complete
	to = 0xffff;
	while(to--){
		// read tunning complete status
		RegVal[20] = akc6951_read_byte(20);
		if(STC) break;		
	}

	// judge tune result
	// if(!to || !TUNED || channel != READ_CHAN) return -1;
	return 0;
}


/*--------------------------------------------------------------------------
	akc6951_AM_tunning
	freq KHz 522-1620
--------------------------------------------------------------------------*/

int akc6951_AM_tunning(long freq)
{
	short int   channel;
	unsigned short int   to;

	printf("akc6951: AM tuning, freq=%ld\n", freq);

	// set request channel
	channel = (short)(freq/3L);	// AM 3KHz Step
	SetChannel(channel);
	akc6951_write_byte_block(2, RegVal+2, 2);	// Register2,Register3

	// triiger tunning
	AMmode();     // select AM
	ResetTune();
	akc6951_write_byte(0, RegVal[0]);	// Register0
	TrigTune();
	akc6951_write_byte(0, RegVal[0]);	// Register0

	// wait for tunning complete
	to = 0xffff;
	while(to--){
		// read tunning complete status
		RegVal[20] = akc6951_read_byte(20);
		if(STC) break;		
	}

	// judge tune result
	// if(!to || !TUNED || channel != READ_CHAN) return -1;
	return 0;
}


/*--------------------------------------------------------------------------
	akc6951_read_byte
--------------------------------------------------------------------------*/

static
char  akc6951_read_byte(char regadr)
{
	char data;

	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(regadr);
	i2cStopClose();

	i2cOpenStart();
	i2cSend(DEVADDR +MR);
	data = i2cRec();
	i2cNAckSend();
	i2cStopClose();

	return data;
}

/*--------------------------------------------------------------------------
	akc6951_read_byte_block	
--------------------------------------------------------------------------*/

static
void  akc6951_read_byte_block(char regadr, char *data, int nbytes)
{
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(regadr);
	i2cStopClose();

	i2cOpenStart();
	i2cSend(DEVADDR +MR);
	while(nbytes--){
		*data++ = i2cRec();
		if(nbytes)
			i2cAckSend();
		else
			i2cNAckSend();
	}
	i2cStopClose();
}


/*--------------------------------------------------------------------------
	akc6951_write_byte
--------------------------------------------------------------------------*/

static
void  akc6951_write_byte(char regadr, char data)
{
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(regadr);
	i2cSend(data);
	i2cStopClose();
	delay_ms(1);
}	


/*--------------------------------------------------------------------------
	akc6951_write_byte_block() 
--------------------------------------------------------------------------*/

static
void  akc6951_write_byte_block(char regadr, char *data, int nbytes)
{
	i2cOpenStart();
	i2cSend(DEVADDR +MW);
	i2cSend(regadr);
	while(nbytes--){
		i2cSend(*data++);
	}
	i2cStopClose();
	delay_ms(1);
}	



/*==========================================================================
	I2C Handling module
	SBDBT32 used I2C3
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

/*--------------------------------------------------------------------------
	I2C master Open and Start
--------------------------------------------------------------------------*/

static
void	i2cOpenStart(void)
{
	// enabled I2C
	I2C3CONbits.I2CEN = 1;
	//delay_us(5);
	delay_ms(1);

	// START 
	I2C3CONbits.SEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.SEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master ReStart
--------------------------------------------------------------------------*/
#if 0
static
void	i2cReStart(void)
{
	// RESTART 
	I2C3CONbits.RSEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.RSEN) break;
	}
}
#endif

/*--------------------------------------------------------------------------
	I2C master Stop and Close
--------------------------------------------------------------------------*/

static
void	i2cStopClose(void)
{
	// STOP 
	I2C3CONbits.PEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.PEN)break;
	}
	//delay_us(5);
	delay_ms(1);

	// disabled I2C
	I2C3CONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C master data send
--------------------------------------------------------------------------*/

static
int	i2cSend(char data)
{
	I2C3TRN = data;

    if(I2C3STATbits.IWCOL){	// If write collision occurs,return -2
		return -2;
    }

	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3STATbits.TRSTAT) break;  // wait until write cycle is complete 
	}

	if (I2C3STATbits.ACKSTAT){ 	// test for ACK condition received
		return -1;	// NACK received
	}

	return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C	master data receive
--------------------------------------------------------------------------*/

static
char	i2cRec(void)
{
	// enabled Receive
	I2C3CONbits.RCEN = 1;

	// wait until receive complete and data stored to I2CRCV register
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.RCEN && I2C3STATbits.RBF) break;
	}

	// clear overflow bit
	I2C3STATbits.I2COV = 0;
    
	return (char)I2C3RCV;
}


/*--------------------------------------------------------------------------
	I2C master Ack send
--------------------------------------------------------------------------*/

static
void	i2cAckSend(void)
{
	// Ack Send
	I2C3CONbits.ACKDT = 0;
	I2C3CONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.ACKEN)break;
	}
}


/*--------------------------------------------------------------------------
	I2C master NotAck send
--------------------------------------------------------------------------*/

static
void	i2cNAckSend(void)
{
	// NAck Send
	I2C3CONbits.ACKDT = 1;
	I2C3CONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2C3CONbits.ACKEN) break;
	}
}


/*** end of akc6951.c ******************************************************/
