/***********************************************************************

	rtc.c
	real time clock, and timer module

	2013/7/24  S.Suwa http://www.suwa-koubou.jp	
***********************************************************************/

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include <stdio.h>
#include <stdlib.h>

#include "rtc.h"

void	rtc_init(void);
void	rtc_set(char *date_time);
char   *get_rtcstr(char *buf);
void	get_rtc(char data[]);
DWORD	get_fattime (void);

static	volatile DWORD  timer_tick;
static	volatile WORD   year;
static	volatile BYTE	leap, month, day, hour, minute, second, wday;
static	const    BYTE   mdays[] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31,}; 

DWORD	volatile AplTimer1;	// 1msec interval down counter1 for Application
DWORD	volatile AplTimer2;  // 1msec interval down counter2 for Application
int     volatile UpdateMinutes; // 1minutes update flg
int		volatile isRtcAdjustTime;	// xx:59:45 
int     volatile isTimeAdjusted; // time adjusted flag


/*----------------------------------------------------------------------
	initialize timer module
----------------------------------------------------------------------*/

void rtc_init(void)
{
	IEC0SET = _IEC0_T1IE_MASK; // enable T1 interrupt
	IFS0CLR = _IFS0_T1IF_MASK; // clear interrupt flag

	IPC1bits.T1IP = 2;

	//--- set interval@1ms timer
	T1CON = 0x0010;	// timer(T1), set prescale 1:8
	TMR1 = 0;
	PR1 = GetPeripheralClock()/8/1000L;	// period 1msec 
	T1CONSET = _T1CON_TON_MASK;	// start T1

	set_rtcstr("2013/08/13(2) 09:00:00");
}


/*----------------------------------------------------------------------
	timer set             1         2
                0123456789012345678901
	set_rtcstr("2013/07/16(2) 09:00:00");
----------------------------------------------------------------------*/

void	set_rtcstr(char *date_time)
{
	int	y,m,d,w,hh,mm,ss;

	// reset time adjusted flag
	isTimeAdjusted = 0;

	y = atoi(date_time);
	m = atoi(date_time+5);
	d = atoi(date_time+8);
	w = atoi(date_time+11);
	hh = atoi(date_time+14);
	mm = atoi(date_time+17); 
	ss = atoi(date_time+20);

	leap = (y % 4 == 0) && (y % 100 != 0 || y % 400 == 0);
	lock_timer(); // disable T1 interrupt
	year = y; month = m; day = d; wday = w;
	hour = hh; minute = mm; second = ss;
	unlock_timer(); // enable T1 interrupt

}


/*----------------------------------------------------------------------
	get_rtcstr(buf)
	 write to buf, "2012/09/01 07:52:05"
----------------------------------------------------------------------*/

char *get_rtcstr(char *buf)
{
	lock_timer(); // disable T1 interrupt
	sprintf(buf, "%4d/%02d/%02d(%d) %02d:%02d:%02d",
		year, month, day, wday, hour, minute, second);
	unlock_timer(); // enable T1 interrupt
	return buf;
}

/*----------------------------------------------------------------------
	set_rtc(data[])
	 write time data from data[] array
----------------------------------------------------------------------*/

void	set_rtc(char data[])
{
	lock_timer(); // disable T1 interrupt
	year   = data[0]+2000;
	month  = data[1];
	day    = data[2];
	hour   = data[3];
	minute = data[4];
	second = data[5];
	wday   = data[6];
	timer_tick = 0;
	unlock_timer(); // enable T1 interrupt
}


/*----------------------------------------------------------------------
	get_rtc(data[])
	 write time data to data[] array
----------------------------------------------------------------------*/

void	get_rtc(char data[])
{
	lock_timer(); // disable T1 interrupt
	data[0] = year -2000;
	data[1] = month;
	data[2] = day;
	data[3] = hour;
	data[4] = minute;
	data[5] = second;
	data[6] = wday;
	unlock_timer(); // enable T1 interrupt
}

/*----------------------------------------------------------------------
	Timer T1 1msec interval timer.
----------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl2), vector(_TIMER_1_VECTOR) )) _T1Interrupt(void)
{
	// clear interrupt flag
	clear_timer();

	// provide timer
	if(AplTimer1) AplTimer1--;
	if(AplTimer2) AplTimer2--;

	// disk i/o command timeout timer
	disk_timerproc();

	// real time clock count up
	if(++timer_tick >= 1000){
	  timer_tick = 0;
      // power monitor LED blink 1sec interval
	  if(!isTimeAdjusted) pwr_led_blink();
	  if(++second >= 60){
	    second = 0;
		if(!UpdateMinutes) UpdateMinutes = 1;
	    if(++minute >= 60){
	      minute = 0;
	      if(++hour >= 24){
	        hour = 0;
			if(++wday > 6) wday = 0;
	        if(++day > (month == 2 && leap ? 29: mdays[month-1]) ){         
	          day = 1;
	          if(++month > 12){
	            month = 1;
	            year++;
				leap = (year % 4 == 0) && (year % 100 != 0 || year % 400 == 0);
	          } // month
	        } // day
	      } // hour
	    } // minute
	  } // second
	  if(second == 45 && minute == 59 && !isRtcAdjustTime){
        isRtcAdjustTime = 1;
		// isTimeAdjusted = 0;
      }   
	} // tick

}



/*----------------------------------------------------------------------
	adjust time
	data[] = YYMMDD WDAY HH:59:45
	adjust next hour and 00 minute 01 second
----------------------------------------------------------------------*/

void adjust_time(char data[])
{
	lock_timer(); // disable T1 interrupt

	year   = data[0]+2000;
	month  = data[1];
	day    = data[2];
	hour   = data[3];
	minute = 0;
	second = 1;
	wday = data[6];


	if(++hour >= 24){
		hour = 0;
		if(++wday > 6) wday = 0;		
		if(++day > (month == 2 && leap ? 29: mdays[month-1]) ){
			day = 1;
			if(++month > 12){
				month = 1;
				year++;
				leap = (year % 4 == 0) && (year % 100 != 0 || year % 400 == 0);
			} // month
		} // day
	} // hour

	timer_tick = 0;
	if(!UpdateMinutes) UpdateMinutes = 1;
	unlock_timer(); // enable T1 interrupt
}



/*---------------------------------------------------------*/
/* User Provided RTC Function for FatFs module             */
/*---------------------------------------------------------*/
/* This is a real time clock service to be called from     */
/* FatFs module. Any valid time must be returned even if   */
/* the system does not support an RTC.                     */
/* This function is not required in read-only cfg.         */

DWORD get_fattime (void)
{
	DWORD tmr;

	lock_timer(); // disable T1 interrupt
	/* Pack date and time into a DWORD variable */
	tmr =	  (((DWORD)year - 1980) << 25)
			| ((DWORD)month << 21)
			| ((DWORD)day << 16)
			| (WORD)(hour << 11)
			| (WORD)(minute << 5)
			| (WORD)(second >> 1);
	unlock_timer(); // enable T1 interrupt

	return tmr;
}


/*** end of rtc.c ****************************************************/
